/*
 *   ORXONOX - the hottest 3D action shooter ever to exist
 *                    > www.orxonox.net <
 *
 *
 *   License notice:
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *   Author:
 *      Gani Aliguzhinov
 *   Co-authors:
 *      ...
 *
 */

#ifndef _ActionpointController_H__
#define _ActionpointController_H__

#include "controllers/FightingController.h"
#include "tools/Timer.h"
#include "tools/interfaces/Tickable.h"
#include "../modules/pickup/PickupSpawner.h"
#include <map>

namespace orxonox
{
    namespace Action
    {  
        enum Value
        {
            NONE, FLY, FIGHT, PROTECT, FIGHTALL, ATTACK
        };
        
    }
    
    struct Point {
        Action::Value action;
        std::string name;
        Vector3 position;
        bool inLoop;
    } ;
    namespace PickupType
    {
        enum Value
        {  
            NONE, DAMAGE, HEALTH, SPEED, PORTAL
        };
    }

    class _OrxonoxExport ActionpointController : public FightingController, public Tickable
    {
        public:
            ActionpointController(Context* context);
            virtual ~ActionpointController();
            virtual void XMLPort(Element& xmlelement, XMLPort::Mode mode);          
               
            virtual void tick(float dt);

            void addActionpoint(WorldEntity* waypoint);
            WorldEntity* getActionpoint(unsigned int index) const;           
            virtual void stayNearProtect();
            virtual void action(); //<! action() is called in regular intervals managing the bot's behaviour.
            virtual void takeActionpoints (const std::vector<Point>& vector, const std::vector<Point>& loop, bool b);

            virtual Action::Value getAction ();
            virtual std::string getActionName();

            void setAction (Action::Value action);
            void setAction (Action::Value action, ControllableEntity* target);
            void setAction (Action::Value action, const Vector3& target);
            void setAction (Action::Value action, const Vector3& target,  const Quaternion& orient );

            virtual bool setWingman(ActionpointController* wingman)
                { return false; }
            virtual bool hasWingman()
                { return true; }
            virtual bool setFollower(ActionpointController* myFollower)
                { return false; }
            virtual bool hasFollower()
                { return true; }
               
        protected:
                void startAttackingEnemiesThatAreClose();
                WeakPtr<ActionpointController> myWingman_;
                WeakPtr<ActionpointController> myFollower_;
                WeakPtr<ActionpointController> myDivisionLeader_;
            //----[Actionpoint information]----
                Action::Value action_;
                std::string protectName_;
                std::string targetName_;
                std::vector<WeakPtr<WorldEntity> > actionpoints_;
                float squaredaccuracy_;
                std::vector<Point > parsedActionpoints_;
                std::vector<Point > loopActionpoints_;
                bool bInLoop_;
                bool bLoop_;   
                bool bEndLoop_;  
                bool bTakenOver_;  
            //----[/Actionpoint information]----
                void setProtect (ControllableEntity* protect);
                ControllableEntity* getProtect (); 
                WeakPtr<ControllableEntity> protect_;     
                void fillLoop();
                void fillLoopReversed();
                void moveBackToTop();
                void setClosestTarget();
                Pawn* closestTarget();
            //----[Actionpoint methods]----
                void executeActionpoint();
                void nextActionpoint();
                void pushAction(Point& p);
                void popAction();
            //----[Actionpoint methods]----          
                bool bFirstTick_; 

                void setSpawners();
                Vector3 bestHealthPickup (float searchRadius);
                // Point closestPickup(PickupType::Value pickupType);
                std::multimap <int, std::pair<PickupSpawner*, bool> > healthSpawners_; 
                std::multimap <int, std::pair<PickupSpawner*, bool> > damageSpawners_; 
                std::multimap <int, std::pair<PickupSpawner*, bool> > speedSpawners_; 

        private:
            
    };
}

#endif /* _ActionpointController_H__ */
