/*
 *   ORXONOX - the hottest 3D action shooter ever to exist
 *                    > www.orxonox.net <
 *
 *
 *   License notice:
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *   Author:
 *      Gani Aliguzhinov
 *   Co-authors:
 *      Fabian 'x3n' Landau, Dominik Solenicki
 *
 */

#ifndef _FightingController_H__
#define _FightingController_H__


#include "controllers/FlyingController.h"

namespace orxonox
{
    class _OrxonoxExport FightingController : public FlyingController
    {

        public:
            FightingController(Context* context);
            virtual ~FightingController();
            
            virtual void XMLPort(Element& xmlelement, XMLPort::Mode mode);
       
            float squaredDistanceToTarget() const;
            bool isLookingAtTarget(float angle) const;
            bool hasTarget() const;
            ControllableEntity* getTarget() const
                { return this->target_; }
            bool bKeepFormation_;
           
        protected:
            void setTarget(ControllableEntity* target);

            void setPositionOfTarget(const Vector3& target);
            void setOrientationOfTarget(const Quaternion& orient);
            void stopLookingAtTarget();
            void startLookingAtTarget();
            void lookAtTarget(float dt);

            void maneuver();
            void dodge(const Vector3& thisPosition, Vector3& diffUnit);
            void dodgeTowards (Vector3& position);
            bool canFire();
            void doFire();
            WeakPtr<ControllableEntity> target_;
            void setClosestTarget();
         
            bool bHasPositionOfTarget_;
            Vector3 positionOfTarget_;
            bool bHasOrientationOfTarget_;
            Quaternion orientationOfTarget_;
            Pawn* closestTarget() const;

            int attackRange_;
            bool bShooting_;
            int maneuverCounter_;   
            bool bLookAtTarget_; 
            float deltaHp;
            float previousHp;
            //WEAPONSYSTEM DATA
            int rocketsLeft_;
            float timeout_;
            bool bFiredRocket_;
            std::map<std::string, int> weaponModes_; //<! Links each "weapon" to it's weaponmode - managed by setupWeapons()
            //std::vector<int> projectiles_; //<! Displays amount of projectiles of each weapon. - managed by setupWeapons()
            void setupWeapons(); //<! Defines which weapons are available for a bot. Is recalled whenever a bot was killed.
            bool bSetupWorked; //<! If false, setupWeapons() is called.
            int getFiremode(std::string name);
          
    };
}

#endif /* _FightingController_H__ */
